from six.moves.urllib.parse import urlparse
from ..errors import InvalidURL


class URL(object):
    """Validate that value is a well formed URL"""
    def __init__(self, schemes=None):
        """
        Args:
            schemes (set): a set of allowed schemes
        """
        self._schemes = schemes

    def __call__(self, value, path):
        """Validate URL format of value

        Args:
            value (object): value to validate
            path (list): error path
        """
        result = urlparse(value)
        if self._schemes and result.scheme not in self._schemes or\
                not (result.scheme and result.netloc):
            raise InvalidURL(value, path=path)



import pytest

from lupin import fields as f, errors as e, validators as v


@pytest.fixture
def field():
    return f.Field(validators=[v.Type(str)])


class TestDump(object):
    def test_returns_the_same_value(self, field):
        assert field.dump("46") == "46"


class TestLoad(object):
    def test_returns_the_same_value(self, field):
        assert field.load("46") == "46"


class TestExtractAttr(object):
    def test_returns_attr_from_object(self, field, thief):
        value = field.extract_attr(thief, "last_name")
        assert value == thief.last_name


class TestValidate(object):
    def test_raise_error_if_invalid_data(self, field):
        with pytest.raises(e.InvalidType):
            field.validate(46, [])



import pytest

from lupin.fields import Constant
from lupin.errors import NotEqual


@pytest.fixture
def field():
    return Constant(46)


class TestLoad(object):
    def test_returns_fixed_value(self, field):
        assert field.load(None) == 46


class TestDump(object):
    def test_returns_fixed_value(self, field):
        assert field.dump(None) == 46


class TestExtractAttr(object):
    def test_returns_fixed_value(self, field):
        assert field.extract_attr(None) == 46


class TestValidate(object):
    def test_raises_error_if_not_constant_value(self, field):
        with pytest.raises(NotEqual):
            field.validate("arsene", [])

    def test_does_nothing_if_constant_value(self, field):
        field.validate(46, [])



# encoding: utf-8
import pytest

from lupin import Mapper
from lupin.errors import InvalidDocument
from tests.fixtures import Thief


@pytest.fixture
def mapping(thief_schema):
    mapper = Mapper()
    return mapper.register(Thief, thief_schema)


class TestLoad(object):
    def test_returns_a_thief(self, mapping, thief_data):
        thief = mapping.load(thief_data)
        assert isinstance(thief, Thief)
        assert thief.last_name == "Lupin"
        assert thief.first_name == "Arsène"


class TestDump(object):
    def test_returns_a_dictionary(self, mapping, thief, thief_data):
        data = mapping.dump(thief)
        assert data == thief_data


class TestValidate(object):
    def test_raise_exception_if_invalid_data(self, mapping, thief_data):
        thief_data["firstName"] = 46
        with pytest.raises(InvalidDocument):
            mapping.validate(thief_data)

    def test_does_nothing_if_valid(self, mapping, thief_data):
        mapping.validate(thief_data)


class TestCanHandle(object):
    def test_returns_false_if_wrong_class(self, mapping, diamond):
        assert not mapping.can_handle(diamond)

    def test_returns_true_if_right_class(self, mapping, thief):
        assert mapping.can_handle(thief)



from ..errors import InvalidIn


class In(object):
    """Validates that a value is contained in a set of values"""

    def __init__(self, values):
        """
        Args:
            values (set|list|tuple): expected values
        """
        self._values = values

    def __call__(self, value, path):
        """Validate that value is contained in the expected ones

        Args:
            value (object): value to validate
            path (list): error path
        """
        if value not in self._values:
            raise InvalidIn(value, self._values, path)



# coding: utf-8
import pytest

from lupin import Mapper, Mapping, Schema, String
from lupin.errors import MissingMapping
from tests.fixtures import Thief


@pytest.fixture
def mapper():
    return Mapper()


@pytest.fixture(autouse=True)
def mapping(mapper, thief_schema):
    return mapper.register(Thief, thief_schema)


class TestRegister(object):
    def test_returns_a_mapping(self, mapper, thief_schema):
        mapping = mapper.register(Thief, thief_schema)
        assert isinstance(mapping, Mapping)

    def test_sets_the_default_schema(self, mapper, thief_schema, thief):
        mapper.register(Thief, thief_schema)
        other_schema = Schema({
            "firstName": String(binding="first_name")
        })
        mapper.register(Thief, other_schema, default=True)
        data = mapper.dump(thief)
        assert data == {
            "firstName": "Arsène"
        }

    def test_handles_multiple_classes(self, mapper, thief_schema, thief_data):
        OtherThiefClass = type("OtherThiefClass", (Thief,), {})
        mapper.register((Thief, OtherThiefClass), thief_schema)
        thief = OtherThiefClass("Arsène", "Lupin")
        data = mapper.dump(thief)
        assert data == thief_data


class TestDump(object):
    def test_returns_thief_data(self, thief, mapper, thief_data, mapping):
        data = mapper.dump(thief, mapping)
        assert data == thief_data

    def test_uses_default_mapping(self, mapper, thief, thief_data):
        data = mapper.dump(thief)
        assert data == thief_data

    def test_dumps_list_of_objects(self, mapper, thief, thief_data):
        data = mapper.dump([thief])
        assert data == [thief_data]

    def test_raises_exception_if_no_mapping(self, mapper):
        with pytest.raises(MissingMapping):
            mapper.dump(46)


class TestLoad(object):
    def test_returns_a_thief(self, mapper, thief_data, mapping):
        thief = mapper.load(thief_data, mapping)
        assert isinstance(thief, Thief)

    def test_loads_list(self, mapper, thief_data, mapping):
        thieves = mapper.load([thief_data], mapping)
        assert isinstance(thieves, list)
        assert isinstance(thieves[0], Thief)


class TestLoadAttrs(object):
    def test_returns_a_dict(self, mapper, thief_data, mapping):
        attrs = mapper.load_attrs(thief_data, mapping)
        assert attrs == {
            "first_name": "Arsène",
            "last_name": "Lupin"
        }

    def test_loads_list(self, mapper, thief_data, mapping):
        thieves = mapper.load_attrs([thief_data], mapping)
        assert isinstance(thieves, list)
        assert isinstance(thieves[0], dict)



from .type import Type  # NOQA
from .equal import Equal  # NOQA
from .match import Match  # NOQA
from .in_validator import In  # NOQA
from .length import Length  # NOQA
from .datetime_format import DateTimeFormat  # NOQA
from .between import Between  # NOQA
from .url import URL  # NOQA



import pytest

from lupin.errors import InvalidLength
from lupin.validators import Length


@pytest.fixture
def validator():
    return Length(min=2, max=3)


class TestCall(object):
    def test_raise_error_if_not_enough_values(self, validator):
        with pytest.raises(InvalidLength):
            validator([1], [])

    def test_raise_error_if_to_much_values(self, validator):
        with pytest.raises(InvalidLength):
            validator([1, 2, 3, 4], [])

    def test_does_nothing_if_enough_values(self, validator):
        validator([1, 2], [])



class MissingMapping(Exception):
    """Error raised when no mapper has been defined for a class
    while dumping an instance of that class.
    """
    def __init__(self, type):
        """
        Args:
            type (type): a type
        """
        message = "Missing mapping for object of type %s" % type
        super(MissingMapping, self).__init__(message)
        self.type = type


class InvalidDocument(Exception):
    """Error raised when validating a document.
    It's composed of all the errors detected.
    """
    def __init__(self, errors):
        super(InvalidDocument, self).__init__()
        self.errors = errors

    def __len__(self):
        """Returns errors count

        Returns:
            int
        """
        return len(self.errors)

    def __getitem__(self, index):
        """Returns error at index

        Args:
            index (int): index

        Returns:
            Exception
        """
        return self.errors[index]


class ValidationError(Exception):
    """Base class for validation errors"""

    def __init__(self, message, path):
        """
        Args:
            message (str): error message
            path (list): path of the invalid data
        """
        super(ValidationError, self).__init__(message)
        self.path = path


class InvalidType(ValidationError):
    """Error raised by `Type` validator"""

    def __init__(self, invalid, expected, path):
        """
        Args:
            invalid (type): invalid type received
            expected (type): type expected
            path (list): path of the invalid data
        """
        message = "Invalid type, got %s instead of %s" % (invalid, expected)
        super(InvalidType, self).__init__(message, path)
        self.invalid = invalid
        self.expected = expected


class NotEqual(ValidationError):
    """Error raised by `Equal` validator"""

    def __init__(self, invalid, expected, path):
        """
        Args:
            invalid (object): invalid value
            expected (object): expected value
            path (list): path of the invalid data
        """
        message = "Invalid value, got %s instead of %s" % (invalid, expected)
        super(NotEqual, self).__init__(message, path)
        self.invalid = invalid
        self.expected = expected


class InvalidMatch(ValidationError):
    """Error raised by `Match` validator"""

    def __init__(self, invalid, regexp, path):
        """
        Args:
            invalid (str): invalid value
            regexp (regexp): a regexp
            path (list): path of the invalid data
        """
        message = "Value \"%s\" does not match pattern \"%s\"" % (invalid, regexp.pattern)
        super(InvalidMatch, self).__init__(message, path)
        self.invalid = invalid
        self.regexp = regexp


class InvalidIn(ValidationError):
    """Error raised by `In` validator"""

    def __init__(self, invalid, expected, path):
        """
        Args:
            invalid (str): invalid value
            expected (list): list of expected values
            path (list): path of the invalid data
        """
        message = "Value \"%s\" is not in %s" % (invalid, expected)
        super(InvalidIn, self).__init__(message, path)
        self.invalid = invalid
        self.expected = expected


class InvalidLength(ValidationError):
    """Error raised by `Length` validator"""

    def __init__(self, length, min, max, path):
        """
        Args:
            length (int): received length
            min (int): minimum length
            max (int): maximum length
            path (list): path of the invalid data
        """
        message = "Got %s items, should have been between %s and %s" % (length, min, max)
        super(InvalidLength, self).__init__(message, path)
        self.max = max
        self.min = min
        self.length = length


class InvalidRange(ValidationError):
    """Error raised by `Between` validator"""

    def __init__(self, value, min, max, path):
        """
        Args:
            value (int): value received
            min (int): minimum value
            max (int): maximum value
            path (list): path of the invalid data
        """
        message = "%s is not between %s and %s" % (value, min, max)
        super(InvalidRange, self).__init__(message, path)
        self.max = max
        self.min = min
        self.value = value


class InvalidURL(ValidationError):
    """Error raised by `URL` validator"""

    def __init__(self, invalid, path):
        """
        Args:
            invalid (str): invalid URL
            path (list): path of the invalid data
        """
        message = "Invalid URL : %s" % (invalid)
        super(InvalidURL, self).__init__(message, path)
        self.invalid = invalid


class InvalidDateTimeFormat(ValidationError):
    """Error raised by `DateTimeFormat` validator"""

    def __init__(self, value, format, path):
        """
        Args:
            value (str): invalid datetime string
            format (str): format used to parse datetime
            path (list): path of the invalid data
        """
        message = "Date value \"%s\" can't be parsed with format \"%s\"" % (value, format)
        super(InvalidDateTimeFormat, self).__init__(message, path)
        self.value = value
        self.format = format


class MissingPolymorphicKey(ValidationError):
    """Error raised if Polymorphic object do not contain a type key"""

    def __init__(self, key, path):
        """
        Args:
            key (str): polymorphic key
            path (list): path of the invalid data
        """
        message = "Polymorphic document does not contain the \"%s\" key " % key
        super(MissingPolymorphicKey, self).__init__(message, path)
        self.key = key


class InvalidPolymorphicType(ValidationError):
    """Error raised by a polymorphic field when it does not have a
    mapping for the data it tries to validate
    """
    def __init__(self, invalid_type, supported_types, path):
        """
        Args:
            invalid_type (str): invalid type received
            supported_types (list): valid types supported
            path (list): path of the invalid data
        """
        message = "Invalid polymorphic document  \"%s\" is not supported only \"%s\" " % (invalid_type, supported_types)
        super(InvalidPolymorphicType, self).__init__(message, path)
        self.invalid_type = invalid_type
        self.supported_types = supported_types


class MissingKey(ValidationError):
    """Error raised when a key is missing from data"""

    def __init__(self, key, path):
        """
        Args:
            key (str): missing key
            path (list): path of the missing data
        """
        message = "Document do not containe the \"%s\" key" % key
        super(MissingKey, self).__init__(message, path)
        self.key = key



from . import Field
from ..validators import Type
from ..errors import MissingPolymorphicKey, InvalidPolymorphicType
from ..utils import get_mapping


class PolymorphicObject(Field):
    """Field used to handle a key that can contain multiple object types

    Example:
        PolymorphicObject(on="type",
                          mappings={
                              "diamond": diamond_mapping,
                              "painting": painting_mapping
                          })
    """
    def __init__(self, on, mappings, **kwargs):
        """
        Args:
            on (str): JSON key used to get the object type
            mappings (dict): mapping used for each values used for the `on` key
        """
        kwargs.setdefault("validators", []).append(Type(dict))
        super(PolymorphicObject, self).__init__(**kwargs)
        self._on = on
        self._mappings_by_json_value = mappings
        self._mappings_by_type = {cls: mapping
                                  for mapping in mappings.values()
                                  for cls in mapping.classes}

    def load(self, value):
        """Loads python objects from JSON object

        Args:
            value (dict): JSON data

        Returns:
            object
        """
        if value is None:
            return None
        return self._mappings_by_json_value[value[self._on]].load(value)

    def dump(self, value):
        """Dump object

        Args:
            value (object): objects to dump

        Returns:
            dict
        """
        mapping = get_mapping(self._mappings_by_type, value)
        return mapping.dump(value)

    def validate(self, value, path):
        """Validate each items of list against nested mapping.

        Args:
            value (list): value to validate
            path (list): JSON path of value
        """
        super(PolymorphicObject, self).validate(value, path)
        if self._on not in value:
            raise MissingPolymorphicKey(self._on, path)

        obj_type = value[self._on]
        if obj_type not in self._mappings_by_json_value:
            raise InvalidPolymorphicType(invalid_type=obj_type,
                                         supported_types=list(self._mappings_by_json_value.keys()),
                                         path=path+[self._on])

        mapping = self._mappings_by_json_value[value[self._on]]
        mapping.validate(value, path=path)



import pytest

from lupin.validators import Type
from lupin.errors import InvalidType


@pytest.fixture
def validator():
    return Type(int)


class TestCall(object):
    def test_raise_error_if_invalid_type(self, validator):
        with pytest.raises(InvalidType):
            validator("not an int", [])

    def test_does_nothing_if_valid(self, validator):
        validator(1, [])



import pytest
from lupin.validators import URL
from lupin.errors import InvalidURL


VALID_URLS = [
    "http://www.example.com",
    "http://www.example.com/logo.png",
    "http://www.example.com/logo.png?var1=1",
    "http://1.1.1.1:80/logo.png?var1=1"
]

INVALID_URLS = [
    "www.example.com",
    "http://"
]


@pytest.fixture
def validator():
    return URL(schemes={"http"})


class TestCall(object):
    @pytest.mark.parametrize("url", VALID_URLS)
    def test_does_not_raise_error_if_ok(self, validator, url):
        validator(url, path=[])

    def test_raise_error_if_invalid_scheme(self, validator):
        with pytest.raises(InvalidURL):
            validator("ftp://1.1.1.1/", path=[])

    @pytest.mark.parametrize("url", INVALID_URLS)
    def test_raise_error_if_invalid_url(self, validator, url):
        with pytest.raises(InvalidURL):
            validator(url, path=[])



import pytest
from lupin.utils import get_mapping
from lupin.errors import MissingMapping
from tests.fixtures import Painting, Jewel


class OilPainting(Painting):
    pass


class TestGetPolymorphicMapping(object):
    def test_raise_error_if_no_mapping_found(self, mona_lisa):
        with pytest.raises(MissingMapping):
            get_mapping({}, mona_lisa)

    def test_returns_mapping_of_object(self, mona_lisa, painting_mapping, jewel_mapping):
        mapping = get_mapping({Jewel: jewel_mapping, Painting: painting_mapping},
                              mona_lisa)
        assert mapping == painting_mapping

    def test_returns_mapping_of_parent_class_if(self, painting_mapping, jewel_mapping):
        painting = OilPainting(None, None)
        mapping = get_mapping({Jewel: jewel_mapping, Painting: painting_mapping},
                              painting)
        assert mapping == painting_mapping



import pytest

from lupin.fields import Dict
from lupin.errors import InvalidType


@pytest.fixture
def field():
    return Dict()


class TestValidate(object):
    def test_raises_error_if_not_a_dict(self, field):
        with pytest.raises(InvalidType):
            field.validate([], [])

    def test_does_nothing_if_dict(self, field):
        field.validate({}, [])



from . import Field
from ..validators import Type


class Int(Field):
    """Field used to handle int values"""

    def __init__(self, **kwargs):
        kwargs.setdefault("validators", []).append(Type(int))
        super(Int, self).__init__(**kwargs)



from collections import defaultdict

from . import Mapping, bind
from .errors import MissingMapping


class Mapper(object):
    """Mapper is used to associate the schemas with their python classes"""

    def __init__(self, default_factory=bind):
        """
        Args:
            default_factory (callable): default factory used to instantiate objects
        """
        self._default_factory = default_factory
        self._mappings = defaultdict(list)

    def register(self, cls, schema, default=False, factory=None):
        """Associate a class with a schema

        If cls is a tuple then it registers all classes for the same schema.
        But the resulting mapping will instantiate the first class from the tuple on .load()

        Args:
            cls (class): a python class or a tuple of class to register multiple classes with the same schema
            schema (Schema): schema that will be used to dump & load objects of klass
            default (bool): if True, this schema will be the default one for dumping instances of `cls`
            factory (callable): factory method used to instantiate objects when loading from JSON

        Returns:
            Mapping
        """
        if not isinstance(cls, tuple):
            cls = (cls,)
        factory = factory or self._default_factory
        mapping = Mapping(cls, schema, factory)
        for klass in cls:
            mapping_index = 0 if default else len(self._mappings[klass])
            self._mappings[klass].insert(mapping_index, mapping)
        return mapping

    def load(self, data, mapping, allow_partial=False):
        """Loads an instance of klass from JSON data

        Args:
            data (dict|list): JSON data
            mapping (Mapping): mapping used to load data
            allow_partial (bool): allow partial schema, won't raise error if missing keys

        returns:
            object
        """
        if isinstance(data, (list, set, tuple)):
            return [self.load(item, mapping, allow_partial) for item in data]

        mapping.validate(data, allow_partial=allow_partial)
        return mapping.load(data, allow_partial=allow_partial)

    def load_attrs(self, data, mapping, allow_partial=False):
        """Loads attributes dictionary from `data`

        Args:
            data (dict): dictionary of data
            allow_partial (bool): allow partial schema, won't raise error if missing keys

        Returns:
            dict
        """
        if isinstance(data, (list, set, tuple)):
            return [self.load_attrs(item, mapping, allow_partial) for item in data]

        mapping.validate(data, allow_partial=allow_partial)
        return mapping.load_attrs(data, allow_partial=allow_partial)

    def dump(self, obj, mapping=None):
        """Dump object into its JSON representation.
        If no mapping is provided, the default one will be used.

        Args:
            obj (object|list): object to dump

        Returns:
            dict
        """
        if isinstance(obj, (list, set, tuple)):
            return [self.dump(item, mapping) for item in obj]

        if not mapping:
            object_type = type(obj)
            if object_type not in self._mappings:
                raise MissingMapping(object_type)
            mapping = self._mappings[type(obj)][0]

        return mapping.dump(obj)



from lupin.errors import MissingMapping


def get_mapping(mappings, obj):
    """Get mapping for object
    If no mappings is found then it falls back to look
    for a mapping of a superclass of obj.

    Args:
        mappings (dict): index between object types and their mapping
        obj (object): object to get mapping for

    Returns:
        Mapping
    """
    mapping = mappings.get(type(obj))
    if mapping:
        return mapping
    else:
        mapping_objects = list(mappings.values())
        for mapping in mapping_objects:
            if mapping.can_handle(obj):
                mappings[type(obj)] = mapping
                return mapping

    raise MissingMapping(type(obj))



import pytest

from lupin.fields import String
from lupin.errors import InvalidType


@pytest.fixture
def field():
    return String()


class TestValidate(object):
    def test_raises_error_if_not_a_string(self, field):
        with pytest.raises(InvalidType):
            field.validate(46, [])

    def test_does_nothing_if_string(self, field):
        field.validate("46", [])



import pytest

from lupin import Mapper, fields as f
from lupin.errors import InvalidType, InvalidDocument, MissingPolymorphicKey,\
    InvalidPolymorphicType, MissingMapping

from tests.fixtures import Jewel, Painting


@pytest.fixture
def field(jewel_schema, painting_schema):
    mapper = Mapper()
    jewel_mapping = mapper.register(Jewel, jewel_schema)
    painting_mapping = mapper.register(Painting, painting_schema)
    return f.PolymorphicObject(on="type",
                               mappings={
                                   "jewel": jewel_mapping,
                                   "painting": painting_mapping
                               })


class TestLoad(object):
    def test_loads_a_diamond(self, field, diamond_data):
        diamond = field.load(diamond_data)
        assert isinstance(diamond, Jewel)

    def test_loads_a_painting(self, field, mona_lisa_data):
        painting = field.load(mona_lisa_data)
        assert isinstance(painting, Painting)

    def test_returs_none_if_none_value(self, field):
        assert field.load(None) is None


class TestDump(object):
    def test_dumps_a_painting(self, field, mona_lisa, mona_lisa_data):
        result = field.dump(mona_lisa)
        assert result == mona_lisa_data

    def test_dumps_a_diamond(self, field, diamond, diamond_data):
        result = field.dump(diamond)
        assert result == diamond_data


class TestValidate(object):
    def test_raise_exception_if_not_dict(self, field):
        with pytest.raises(InvalidType):
            field.validate("", [])

    def test_raise_exception_if_object_is_invalid(self, field, diamond_data):
        diamond_data["carat"] = "20"
        with pytest.raises(InvalidDocument) as exc:
            field.validate(diamond_data, [])

        errors = exc.value
        error = errors[0]
        assert error.path == ["carat"]

    def test_raises_exception_if_no_polymorphic_key(self, field, diamond_data):
        del diamond_data["type"]
        with pytest.raises(MissingPolymorphicKey):
            field.validate(diamond_data, [])

    def test_raises_exception_if_invalid_polymorphic_type(self, field, diamond_data):
        diamond_data["type"] = "car"
        with pytest.raises(InvalidPolymorphicType) as err:
            field.validate(diamond_data, [])

    def test_does_nothing_if_all_items_are_valid(self, field, diamond_data):
        field.validate(diamond_data, [])



# coding: utf-8
import pytest

from lupin import Constant, fields as f
from lupin.errors import InvalidDocument, InvalidType, MissingKey
from tests.fixtures import Thief


class TestLoad(object):
    def test_returns_a_thief(self, thief_schema, thief_data):
        thief = thief_schema.load(Thief, thief_data)
        assert isinstance(thief, Thief)
        assert thief.last_name == "Lupin"
        assert thief.first_name == "Arsène"


class TestLoadAttrs(object):
    def test_returns_a_dict_with_all_attrs(self, thief_schema, thief_data):
        attrs = thief_schema.load_attrs(thief_data)
        assert attrs == {
            "first_name": "Arsène",
            "last_name": "Lupin"
        }

    def test_returns_default_value_if_missing_key(self, thief_schema, thief_data):
        del thief_data["firstName"]
        attrs = thief_schema.load_attrs(thief_data)
        assert attrs == {
            "first_name": None,
            "last_name": "Lupin"
        }

    def test_returns_no_attr_if_allow_partial_and_missing_key(self, thief_schema, thief_data):
        del thief_data["firstName"]
        attrs = thief_schema.load_attrs(thief_data, allow_partial=True)
        assert attrs == {
            "last_name": "Lupin"
        }

    def test_do_not_load_read_only_attribute(self, thief_schema, thief_data):
        thief_schema.add_field("lastName", f.String(read_only=True))
        attrs = thief_schema.load_attrs(thief_data, allow_partial=True)
        assert attrs == {
            "first_name": "Arsène"
        }


class TestDump(object):
    def test_returns_a_dictionary(self, thief_schema, thief, thief_data):
        data = thief_schema.dump(thief)
        assert data == thief_data


class TestAddField(object):
    def test_add_new_field_to_schema(self, thief_schema, thief, thief_data):
        thief_schema.add_field("age", Constant(28))
        result = thief_schema.dump(thief)
        thief_data["age"] = 28
        assert result == thief_data


class TestValidate(object):
    def test_raise_exception_if_invalid_data(self, thief_schema, thief_data):
        thief_data["firstName"] = 46
        with pytest.raises(InvalidDocument) as exc:
            thief_schema.validate(thief_data)

        errors = exc.value
        assert len(errors) == 1
        error = errors[0]
        assert isinstance(error, InvalidType)
        assert error.path == ["firstName"]

    def test_does_nothing_if_valid(self, thief_schema, thief_data):
        thief_schema.validate(thief_data)

    def test_raise_error_if_missing_key(self, thief_schema, thief_data):
        del thief_data["firstName"]
        with pytest.raises(InvalidDocument) as exc:
            thief_schema.validate(thief_data)

        errors = exc.value
        assert len(errors) == 1
        error = errors[0]
        assert isinstance(error, MissingKey)
        assert error.path == ["firstName"]

    def test_do_not_raise_exception_if_allow_partial_and_missing_key(self, thief_schema, thief_data):
        del thief_data["firstName"]
        thief_schema.validate(thief_data, allow_partial=True)

    def test_do_not_raise_exception_if_optional_field(self, thief_schema, thief_data):
        thief_schema.add_field("age", f.String(optional=True))
        assert "age" not in thief_data
        thief_schema.validate(thief_data)



from ..errors import InvalidLength


class Length(object):
    """Validates the size of value"""

    def __init__(self, min=None, max=None):
        """
        Args:
            min (int): minimum length
            max (int): maximum length
        """
        self._min = min
        self._max = max

    def __call__(self, value, path):
        """Validate that value is contained in the expected ones

        Args:
            value (object): value to validate
            path (list): error path
        """
        length = len(value)
        if self._min is not None and length < self._min or\
                self._max is not None and length > self._max:
            raise InvalidLength(length, self._min, self._max, path)



from datetime import datetime
from lupin import fields
import pytest


@pytest.fixture
def field():
    return fields.DateTime()


@pytest.fixture
def time():
    return datetime(2017, 10, 7, 17, 35, 0)


class TestDump(object):
    def test_returns_datetime_as_string(self, field, time):
        assert field.dump(time) == "2017-10-07T17:35:00"


class TestLoad(object):
    def test_returns_a_datetime(self, field, time):
        assert field.load("2017-10-07T17:35:00") == time



from lupin.errors import InvalidType


class Type(object):
    """Validates that data is of a certain type"""

    def __init__(self, expected_type):
        """
        Args:
            expected_type (type): an object type
        """
        self._expected_type = expected_type

    def __call__(self, value, path):
        """Validate that value is of expected type

        Args:
            value (object): value to validate
            path (list): error path
        """
        if not isinstance(value, self._expected_type):
            raise InvalidType(type(value),
                              self._expected_type,
                              path)
